/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

/* Header file for singly-linked list class templates for integer
 * keyed items and string-names items.
 */

#if ! defined _lists_h_
#define _lists_h_

// define the macro below (at your peril!) to obtain reams of debug
// information on the dyanamics of your list operations.

// #define DEBUG_LISTS

#include <stdio.h>
#include "Basic.h"
#include "Strings.h"

// CLASS GROUP: list template classes

template <class K, class T> class slink ;
template <class K, class T> class slist ;
template <class K, class T> class slist_iterator ;
template <class K, class T> class dlink ;
template <class K, class T> class dlist ;
template <class K, class T> class dlist_iterator ;

/*---------------------------------------------------------------------
CLASS
     slink

     List linkage template - singly-linked.

OVERVIEW
     Provides a link object to connect slist members together in a
     linear, singly-linked structure.

----------------------------------------------------------------------*/

template <class K, class T> class slink {

friend class slist<K, T> ;
friend class slist_iterator<K, T> ;

  slink<K, T>* next ;
  T* e ;
  slink (T* a, slink<K, T>* p) { e = a; next = p; }
};

/*---------------------------------------------------------------------
CLASS
     slist

     Singly-linked list template.

OVERVIEW
     A singly-linked linear list template, with some special additional
     properties. It is instantiated with two template arguments K and T
     representing the key type and the data object class.

DESCRIPTION
     This is a list template which permits items in the list to be
     searched for based on comparisons of keys of a parametric type.
     Each slist is instantiated with two template arguments, K and T.
     The K argument is normally either int or char*, representing
     ordinal lists and string lists respectively.

     Items can be searched for using the find() function, for example.
     This takes a type K argument and iterates through the list
     searching for a matching element. Matches are performed by a
     template matching function notequal(), which compares the key
     field of each list element against the value returned by the
     type K index() function which each list element (of class T)
     must provide as a public member function.
----------------------------------------------------------------------*/

template <class K, class T> class slist {

friend class slist_iterator<K, T> ;

public:

        ///////////////////////////////////////////
        //
        // Prepends a new item at the head of the
        // list. 
        //
  void prepend (T* a) {
    if (last)
      last->next = new slink<K, T> (a, last->next) ;
    else {
      last = new slink<K, T> (a, 0) ;
      last->next = last ;
    }
  }

        ///////////////////////////////////////////
        //
        // Appends a new item at the tail of the
        // list.
        //
  void append (T* a) {
    if (last)
      last = last->next = new slink<K, T> (a, last->next) ;
    else {
      last = new slink<K, T> (a, 0) ;
      last->next = last ;
    }
  }

        ///////////////////////////////////////////
        //
        // Searches for the item in the list with
        // the given key value of (class) type K.
        //
        //  [in]  K  the key of the item to be found
        //  [out] T* ptr to item, or zero if not found
        //
  T* find (K key) {
    slist_iterator<K, T> next(*this) ;
    T* p ;
    while ( (p = next()) && notequal(p->index(), key) ) ;
    return p ;
  }

        ///////////////////////////////////////////
        //
        // Searches for the item in the list with
        // the given key value of (class) type K,
        // which also matches the integer variant
        // given as the second parameter.
        //
        //  [in]  K    the key of the item to be found
        //  [in]  int  the variant of the item to be found
        //  [out] T*   ptr to item, or zero if not found
        //
  T* find (K key, int varnt) {
    slist_iterator<K, T> next(*this) ;
    T* p ;
    while ( (p = next()) && 
	    (notequal(p->index(), key) ||
	    (varnt != p->variant())) ) ;
    return p ;
  }

        ///////////////////////////////////////////
        //
        // Removes the item in the list with
        // the given key value of (class) type K,
        // which also matches the integer variant
        // given as the second parameter.
        //
        //  [in]  K    the key of the item to be found
        //  [in]  int  the variant of the item to be found
        //  [out] int  number of items removed (0,1)
        //
  int remove (K key, int v) {
    slink<K, T>** p = &last ;
    int unmatched = 1 ;
    int unfinished = (*p != 0) ;

    while (unfinished && unmatched) {
      unmatched = ( notequal ((*p)->next->e->index(), key )
		    || (v && (v != (*p)->next->e->variant())) ) ;

      if (unmatched)
	p = &((*p)->next) ;

      unfinished = (*p != last) ;
    }

    if (*p && !unmatched) {
      slink<K, T>* tmp = (*p)->next ;
      (*p)->next = tmp->next ;

      if (tmp == last)
      {
	if (tmp->next == tmp)
	  last = 0 ;
	else
	  last = *p ;
      }

      tmp->next = 0 ; // remove tmp from the chain
      delete tmp ;    // before deleting it
      return 1 ;
    }
    return 0 ;
  }

        ///////////////////////////////////////////
        //
        // Removes the item in the list with
        // the given key value of (class) type K.
        //
        //  [in]  K    the key of the item to be found
        //  [out] int  number of items removed (0,1)
        //
  int remove (K key) {
    slink<K, T>** p = &last ;
    int unmatched = 1 ;
    int unfinished = (*p != 0) ;

    while (unfinished && unmatched) {
      unmatched = notequal ((*p)->next->e->index(), key) ;

      if (unmatched)
	p = &((*p)->next) ;

      unfinished = (*p != last) ;
    }

    if (*p && !unmatched) {
      slink<K, T>* tmp = (*p)->next ;
      (*p)->next = tmp->next ;

      if (tmp == last)
      {
	if (tmp->next == tmp)
	  last = 0 ;
	else
	  last = *p ;
      }

      tmp->next = 0 ; // remove tmp from the chain
      delete tmp ;    // before deleting it
      return 1 ;
    }
    return 0 ;
  }


        ///////////////////////////////////////////
        //
        // Detaches and returns the head item from
        // the list.
        //
        //  [out] T*  ptr to head item returned, or zero
        //
  T* get() {
    if (last != 0) {
      slink<K, T>* f = last->next ;
      T* r = f->e ;
      if (f == last)
	last = 0 ;
      else
	last->next = f->next ;
      delete f ;
      return r ;
    }
    else
      return 0 ;
  }

        ///////////////////////////////////////////
        //
        // Returns the item at the head of the list,
        // or zero if the list is empty.
        //
        // [out] T* head item, or zero
        //
  T* hd () {
    if (last != 0) {
      slink<K, T>* f = last->next ;
      T* r = f->e ;
      return r ;
    }
    else
      return 0 ;
  }

        ///////////////////////////////////////////
        //
        // Returns the item at the tail of the list,
        // or zero if the list is empty.
        //
        // [out] T* tail item, or zero
        //
  T* tail () { return last ? last->e : 0 ; }


        ///////////////////////////////////////////
        //
        // Joins the list given by the parameter to the end of
        // this list, clearing the parameter list elements in
        // so doing.
        //
        // [in] slist<K,T>& list to be appended
        //
  void join (slist<K,T> &tt)
  {
    if (last)
      {
	if (tt.last)
	  {
	    slink<K,T>* hh = last->next ;
	    last->next = tt.last->next ;
	    tt.last->next = hh ;
	  }
      }
    else
      last = tt.last ;
    tt.last = 0 ;
  }

        ///////////////////////////////////////////
        //
        // Dismembers the list, leaving the elements
        // pointed to by the list intact.
        //
  void clear () {
    slink<K, T>* l = last ;
#ifdef DEBUG_LISTS
    printf (" clearing list: 0x%08x\n", (unsigned long)last) ;
#endif
    if (l == 0) return ;
    do {
      slink<K, T>* ll = l ;
#ifdef DEBUG_LISTS
      printf ("   list-clear: [0x%08x]: (0x%08x, 0x%08x)\n", 
	      (unsigned long)ll, (unsigned long)ll->next, (unsigned long)ll->e) ;
#endif
      l = l->next ;
      delete ll ;
    } while (l != last) ;
    last = 0 ;
  }

  // zap() deletes both the list structure and the items pointed to
  // by the list.

        ///////////////////////////////////////////
        //
        // Deletes both the list structure and the objects
        // within the list.
        //
  void zap () {
    slink<K, T>* l = last ;
#ifdef DEBUG_LISTS
    printf (" zapping list: 0x%08x\n", (unsigned long)last) ;
#endif
    if (l == 0) return ;
    do {
      slink<K, T>* ll = l ;
      l = l->next ;
#ifdef DEBUG_LISTS
      printf ("   list-zap: [0x%08x]: (0x%08x, 0x%08x)\n", 
	      (unsigned long)ll, (unsigned long)ll->next, (unsigned long)ll->e) ;
#endif
      delete ll->e ;
      delete ll ;
    } while (l != last) ;
    last = 0 ;
  }

        ///////////////////////////////////////////
        //
        // A rather rudimentary print function for
        // displaying the contents of a list
        // on stdout.
        //
  void print ()
  {
    int done = 0 ;
    slink<K, T>* l = last ;
    printf ("0x%09x -> ", (unsigned long)l ) ;
    while (l && !done)
    {
      printf ("(0x%09x, 0x%09x) ", (unsigned long)l->e, (unsigned long)l->next) ;
      l = l->next ;
      done = (l == last) ;
    }
  }

        ///////////////////////////////////////////
        //
        // Returns a non-zero value if the list is
        // empty, otherwise returns 0.
        //
        // [out] int  list empty predicate
        //
  int empty () { return last == 0 ; }

        ///////////////////////////////////////////
        //
        // Constructor for an empty list.
        //
  slist ()     { last = 0 ; }

        ///////////////////////////////////////////
        //
        // Copy constructor.
        //
  slist (slist<K, T> &sl)
    {
      slist_iterator<K, T> next_item (sl) ;
      T* item ;
      while ( (item = next_item()) != 0 )
	this->append (item) ;
    }

        ///////////////////////////////////////////
        //
        // Returns number of items in the list.
        //
  int length ()
    {
      slist_iterator<K, T> next_item (this) ;
      T* item ;
      int count = 0 ;
      while ( (item = next_item()) != 0)
	++count ;
      return count ;
    }

        ///////////////////////////////////////////
        //
        // Constructor for a singleton list.
        //
        //  [in] T*  pointer to initial list item
        //
  slist (T* a) { last = new slink<K, T> (a, 0) ; last->next = last ; }

        ///////////////////////////////////////////
        //
        // Destructor, but does not destroy list items.
        //
  ~slist ()    { clear() ; }

  int notequal (int x, int y)       { return x != y ; }
  int notequal (char* x, char* y)   { return (y) ? strcmp (x, y) : 0 ; }
  int notequal (float x, float y)   { return x != y ; }
  int notequal (double x, double y) { return x != y ; }

protected:
  slink<K, T>* last ;
};


/*---------------------------------------------------------------------
CLASS
     slist_iterator

     Scans the elements of an slist

OVERVIEW
     An iterator class for scanning through the elements of an slist

----------------------------------------------------------------------*/

template <class K, class T> class slist_iterator {

public:

        ///////////////////////////////////////////
        //
        // Constructor for empty list iterator.
        //
  slist_iterator()  { cs = 0 ; ce = 0 ; mark = 0 ; }

        ///////////////////////////////////////////
        //
        // Constructor for initialised list iterator.
        // Use this to begin iterating over the list
        // given as parameter.
        //
        // The current position and the mark are both
        // set to the first item in the list.
        //
        //  [in] slist&  an slist to be iterated over
        //
  slist_iterator(slist<K, T>& s)  { cs = &s; ce = cs->last; mark = ce ; }

        ///////////////////////////////////////////
        //
        // Constructor for initialised list iterator.
        // Use this to begin iterating over the list
        // given as parameter.
        //
        // The current position and the mark are both
        // set to the first item in the list.
        //
        //  [in] slist*  an slist that you want to iterate
        //
  slist_iterator(slist<K, T>* s)  { cs = s ; ce = cs ? cs->last : 0 ; mark = ce ; }
  
        ///////////////////////////////////////////
        //
        // Re-assignment of an existing list iterator.
        //
        //  [in] slist& an slist to be iterated over
        //
  void operator= (slist<K, T>& s) { cs = &s; ce = cs->last ; mark = ce ; }

        ///////////////////////////////////////////
        //
        // Re-assignment of an existing list iterator.
        //
        //  [in] slist& an slist to be iterated over
        //
  void operator= (slist<K, T>* s)
  {
    if (s) {
      cs = s; ce = mark = cs->last ;
    }
    else {
      cs = 0; ce = 0 ;
    }
  }

        ///////////////////////////////////////////
        //
        // The () operator is the principal method of
        // advancing the iterator through the list.
        // It returns the next item in the list and
        // moves the read pointer past that item.
        //
        //  [out] T* ptr to the next list item, or zero
        //           if end of list is reached.
  T* operator()()
  {
    if (!this) return 0 ;
    if (!this->cs) return 0 ;

    T* ret ;
    ret = ((ce) ? (ce = ce->next)->e : 0 );
    if (cs->last == ce)
	ce = 0 ;
    return ret ;
  }

        ///////////////////////////////////////////
        //
        // Returns a pointer to the list item that is
        // next in sequence to be scanned by the ()
        // operator. This permits a one item lookahead
        // so that the next item can be either iterated
        // over or removed.
        //
  T* head ()
  {
    if (! this || ! cs) return 0 ;
    
    return ce ? ce->next->e : 0 ;
  }

        ///////////////////////////////////////////
        //
        // Removes the next unscanned item in the list.
        // This may cause the list to become empty,
        // and the () operator to subsequently return zero.
        // Calls to this function can be safely interleaved with
        // calls to the () operator and the head() function.
        //
  void remove ()
  {
    if (! this || ! cs || ! ce) return ;

    slink<K,T>* d = ce->next ;
    ce->next = ce->next->next ;
    
    if (d == cs->last)
      {
	cs->last = ce ;
	ce = 0 ;
      }

    if (d == ce)
      cs->last = 0 ;

    delete d ;
  }

        ///////////////////////////////////////////
        //
        // Removes and returns the next unscanned
        // item in the list, or zero if the iterator
        // is currently pointing at the end of the list
        // or if the list is empty.
        //
        // This may cause the list to become empty,
        // and the () operator to subsequently return zero.
        // Calls to this function can be safely interleaved with
        // calls to the () operator and the head() function.
        //
  T* pull ()
  {
    if (! this || ! cs || ! ce) return 0 ;

    slink<K,T>* d = ce->next ;
    ce->next = ce->next->next ;
    
    if (d == cs->last)
      {
	cs->last = ce ;
	ce = 0 ;
      }

    if (d == ce)
      cs->last = 0 ;

    T* v = d->e ;
    delete d ;
    return v ;
  }


        ///////////////////////////////////////////
        //
        // Removes and returns the next item from
        // the "Mark" position, moving the mark
        // position on to the next item.
        // This function returns 0 if the mark is not
        // pointing at a valid list node, or if the mark
        // is pointing at the current iterator position.
        //
        // This may cause the mark to point at the end
        // of the list.
        //
  T* pull_mark ()
  {
    if (! this || ! cs || ! mark || mark == ce) return 0 ;

    slink<K,T>* d = mark->next ;
    mark->next = mark->next->next ;
    
    if (d == cs->last)
      {
	cs->last = mark ;
	mark = 0 ;
      }

    if (d == mark)
      {
	cs->last = ce = 0 ;
      }

    T* v = d->e ;
    delete d ;
    return v ;
  }

        ///////////////////////////////////////////
        //
        // Inserts a new item into the list at a
        // position immediately preceeding the next
        // item to be scanned.
        //
        //  [in] T*  ptr to item to be added
        //
  void insert (T* a)
  {
    if (! this || ! cs) return ;

    if (ce)
      {
	ce->next = new slink<K,T> (a, ce->next) ;
	ce = ce->next ;
      }
    else
      cs->append (a) ;
  }

        ///////////////////////////////////////////
        //
        // Inserts a new item into the list at a
        // position immediately preceeding the next
        // item to be scanned.
        //
        //  [in] T*  ptr to item to be added
        //
  void insert (T& a)
  {
    if (! this || ! cs) return ;

    if (ce)
      {
	ce->next = new slink<K,T> (&a, ce->next) ;
	ce = ce->next ;
      }
    else
      cs->append (&a) ;
  }

        ///////////////////////////////////////////
        //
        // Inserts a new item into the list at a
        // position immediately preceeding the Mark
        //
        //  [in] T*  ptr to item to be added
        //
  void insert_mark (T* a)
  {
    if (! this || ! cs) return ;

    if (mark)
      {
	mark->next = new slink<K,T> (a, mark->next) ;
	mark = mark->next ;
      }
    else
      {
	cs->append (a) ;
	mark = cs->last ;
      }
  }

        ///////////////////////////////////////////
        //
        // Rewinds the list iterator to the start of
        // the list.
        //
  void rewind ()
  {
    ce = cs ? cs->last : 0 ;
  }

        ///////////////////////////////////////////
        //
        // Moves the current list iterator position
        // to the start of the list
        //
  void goto_start () { rewind () ; }

        ///////////////////////////////////////////
        //
        // Moves the current position of the list
        // iterator to the end of the list.
        //
  void goto_end () { ce = 0 ; }

        ///////////////////////////////////////////
        //
        // Moves the current position of the list
        // iterator to point at the given argument node.
        // If the argument is not present in the list
        // the current position is set to the end of the
        // list.
        //
  void goto_node (T* node)
  {
    slink<K, T>* p = cs->last ;
    
    while (p && p->next->e != node)
      {
	p = p->next ;
	if (p == cs->last)
	  p = 0 ;
      }
    ce = p ;
  }

        ///////////////////////////////////////////
        //
        // Remembers the current position of this
        // list iterator in the "Mark" variable.
        //
  void set_mark () { mark = ce ; }


        ///////////////////////////////////////////
        //
        // Searches for the given node in the list
        // and sets the mark at that point if found
        //
  void set_mark (T* node) 
  {
    slink<K, T>* p = cs->last ;
    
    while (p && p->next->e != node)
      p = p->next ;
    
    mark = p ;
  }

        ///////////////////////////////////////////
        //
        // Positions the iterator at the position
        // last given to the "Mark" variable.
        //
  void goto_mark () { ce = mark ; }

        ///////////////////////////////////////////
        //
        // Returns the list item stored at the position
        // given by the "Mark" variable, or 0 if the
        // mark is not pointing at a valid list item, or
        // is effectively at the end of the list.
        //
  T* get_mark () { return mark ? mark->next->e : 0 ; }

        ///////////////////////////////////////////
        //
        // Exchanges the "Mark" variable with the
        // current iterator position.
        //
  void exchange_mark () { slink<K, T>* tmp = mark ; mark = ce ; ce = tmp ; }

        ///////////////////////////////////////////
        //
        // Removes a section of this iterator's list
        // and return a pointer to that new list.
        // The section removed is from the "Mark"
        // position up to, but not including, the
        // current position or the end of the list
        // (whichever is reached first). Thus, if Mark
        // is closer to the end of the list than the
        // current position, the section removed will
        // Be delimited by (Mark, End_Of_List), otherwise
        // it will be delimited by (Mark, Current).
        // If Mark is pointing at the end of the list,
        // or is not pointing to a valid position
        // in the list, then no cut operation occurs, and
        // an empty list is returned.
        //
  slist<K, T>* cut ()
  {
    slist<K, T>* nl = new slist<K, T> () ;
    T* item ;

    while (item = pull_mark() )
      nl->append (item) ;

    return nl ;
  }

        ///////////////////////////////////////////
        //
        // The items in the argument slist are copied
        // into this iterator's list immediately prior
        // to the next item to be scanned (i.e. at the
        // current position).
        //
        // Ownership of the argument list is NOT assumed
        // by paste(), and responsibility for deleting
        // the argument list remains with the caller.
        //
  void paste (slist<K, T>& nl)
  {
    slist_iterator<K, T> next_item (nl) ;
    T* item ;

    while (item = next_item())
      insert (item) ;
  }

        ///////////////////////////////////////////
        //
        // The items in the argument slist are copied
        // into this iterator's list immediately prior
        // to the Mark position.
        //
        // Ownership of the argument list is NOT assumed
        // by paste(), and responsibility for deleting
        // the argument list remains with the caller.
        //
  void paste_mark (slist<K, T>& nl)
  {
    slist_iterator<K, T> next_item (nl) ;
    T* item ;

    while (item = next_item())
      insert_mark (item) ;
  }

private:

  slink<K, T>* ce ;
  slist<K, T>* cs ;
  slink<K, T>* mark ;
};

/*---------------------------------------------------------------------
CLASS
     dlink

     List linkage template - doubly-linked.

OVERVIEW
     Provides a link object to connect dlist members together in a
     linear, doubly-linked structure.

----------------------------------------------------------------------*/

template <class K, class T> class dlink {

friend class dlist<K, T> ;
friend class dlist_iterator<K, T> ;

  dlink (T* a, dlink<K, T> *p, dlink<K, T> *n)
  { 
    e = a ;
    prev = p ;
    next = n ;
  }

  dlink<K, T> *prev, *next ;
  T* e ;
};

/*---------------------------------------------------------------------
CLASS
     dlist

     Doubly-linked list template.

OVERVIEW
     A doubly-linked linear list template, with some special additional
     properties. It is instantiated with two template arguments K and T
     representing the key type and the data object class.

DESCRIPTION
     This is a list template which permits items in the list to be
     searched for based on comparisons of keys of a parametric type.
     Each dlist is instantiated with two template arguments, K and T.
     The K argument is normally either int or char*, representing
     ordinal lists and string lists respectively.

     Items can be searched for using the find() function, for example.
     This takes a type K argument and iterates through the list
     searching for a matching element. Matches are performed by a
     template matching function notequal(), which compares the key
     field of each list element against the value returned by the
     type K index() function which each list element (of class T)
     must provide as a public member function.

     Just to complicate matters, it is possible to have several
     concurrent dlist_iterators iterating over the same dlist.
     As iterators can modify the list, there is the possibility of
     one iterator removing a list item that is currently being
     pointed to by another concurrent iterator. To cope with this,
     each dlist object maintains a pointer to all active iterators
     that are iterating over it at any moment. This takes the form
     of a threaded list of dlist_iterators pointed to from within
     the dlist object. Whenever a dlist_iterator removes an item
     it must scan the list of iterators to adjust the current
     entry and mark pointers so that they do not point at a
     deleted entry.
----------------------------------------------------------------------*/

template <class K, class T> class dlist {

friend class dlist_iterator<K, T> ;

public:

        ///////////////////////////////////////////
        //
        // Prepends a new item at the head of the
        // list. 
        //
  void prepend (T* a) {
    if (first)
      {
	dlink<K,T> *f = new dlink<K, T> (a, first->prev, first) ;
	first->prev->next = f ;
	first->prev = f ;
	first = f ;
      }
    else
      {
	first = new dlink<K, T> (a, 0, 0) ;
	first->next = first->prev = first ;
      }
  }

        ///////////////////////////////////////////
        //
        // Appends a new item at the tail of the
        // list.
        //
  void append (T* a) {
    if (first)
      {
	dlink<K,T> *f = new dlink<K, T> (a, first->prev, first) ;
	first->prev->next = f ;
	first->prev = f ;
      }
    else
      {
	first = new dlink<K, T> (a, 0, 0) ;
	first->next = first->prev = first ;
      }
  }

        ///////////////////////////////////////////
        //
        // Searches for the item in the list with
        // the given key value of (class) type K.
        //
        //  [in]  K  the key of the item to be found
        //  [out] T* ptr to item, or zero if not found
        //
  T* find (K key) {
    dlist_iterator<K, T> next(*this) ;
    T* p ;
    while ( (p = next()) && notequal(p->index(), key) ) ;
    return p ;
  }

        ///////////////////////////////////////////
        //
        // Searches for the item in the list with
        // the given key value of (class) type K,
        // which also matches the integer variant
        // given as the second parameter.
        //
        //  [in]  K    the key of the item to be found
        //  [in]  int  the variant of the item to be found
        //  [out] T*   ptr to item, or zero if not found
        //
  T* find (K key, int varnt) {
    dlist_iterator<K, T> next(*this) ;
    T* p ;
    while ( (p = next()) && 
	    (notequal(p->index(), key) ||
	    (varnt != p->variant())) ) ;
    return p ;
  }

        ///////////////////////////////////////////
        //
        // Removes the item in the list with
        // the given key value of (class) type K,
        // which also matches the integer variant
        // given as the second parameter.
        //
        //  [in]  K    the key of the item to be found
        //  [in]  int  the variant of the item to be found
        //  [out] int  number of items removed (0,1)
        //
  int remove (K key, int v) {
    dlink<K, T>** p = &first ;
    int unmatched = 1 ;
    int unfinished = (*p != 0) ;

    while (unfinished && unmatched) {
      unmatched = ( notequal ((*p)->next->e->index(), key )
		    || (v && (v != (*p)->next->e->variant())) ) ;

      if (unmatched)
	p = &((*p)->next) ;

      unfinished = (*p != first) ;
    }

    if (*p && !unmatched)
      {
	dlink<K, T>* tmp = *p ;
	tmp->next->prev = tmp->prev ;
	tmp->prev->next = tmp->next ;

	if (tmp == first)
	  {
	    if (tmp->next == tmp)
	      first = 0 ;
	    else
	      first = first->next ;
	  }
	
	tmp->next = tmp->prev = 0 ; // remove tmp from the chain
	delete tmp ;                // before deleting it
	return 1 ;
      }
    return 0 ;
  }

        ///////////////////////////////////////////
        //
        // Removes the item in the list with
        // the given key value of (class) type K.
        // The datum removed from the list is not
        // deleted or altered in any way.
        //
        //  [in]  K    the key of the item to be found
        //  [out] int  number of items removed (0,1)
        //
  int remove (K key) {
    dlink<K, T>** p = &first ;
    int unmatched = 1 ;
    int unfinished = (*p != 0) ;

    while (unfinished && unmatched) {
      unmatched = notequal ((*p)->e->index(), key) ;

      if (unmatched)
	p = &((*p)->next) ;

      unfinished = (*p != first) ;
    }

    if (*p && !unmatched)
      {
	dlink<K, T>* tmp = *p ;
	tmp->next->prev = tmp->prev ;
	tmp->prev->next = tmp->next ;

	if (tmp == first)
	  {
	    if (tmp->next == tmp)
	      first = 0 ;
	    else
	      first = first->next ;
	  }
	
	tmp->next = tmp->prev = 0 ; // remove tmp from the chain
	delete tmp ;                // before deleting it
	return 1 ;
      }
    return 0 ;
  }


        ///////////////////////////////////////////
        //
        // Detaches and returns the head item from
        // the list.
        //
        //  [out] T*  ptr to head item returned, or zero
        //
  T* get() {
    if (first) {
      dlink<K, T>* f = first ;
      T* r = f->e ;
      if (f == f->next)
	first = 0 ;
      else
	{
	  first = first->next ;
	  f->prev->next = first ;
	  first->prev = f->prev ;
	}
      delete f ;
      return r ;
    }
    else
      return 0 ;
  }

        ///////////////////////////////////////////
        //
        // Returns the item at the head of the list,
        // or zero if the list is empty.
        //
        // [out] T* head item, or zero
        //
  T* hd () {
    if (first) {
      return first->e ;
    }
    else
      return 0 ;
  }

        ///////////////////////////////////////////
        //
        // Returns the item at the tail of the list,
        // or zero if the list is empty.
        //
        // [out] T* tail item, or zero
        //
  T* tail () { return first ? first->prev->e : 0 ; }


        ///////////////////////////////////////////
        //
        // Joins the list given by the parameter to the end of
        // this list, clearing the parameter list elements in
        // so doing.
        //
        // [in] dlist<K,T>& list to be appended
        //
  void join (dlist<K,T> &tt)
  {
    if (first)
      {
	if (tt.first)
	  {
	    dlink<K,T>* hh = first->prev ;
	    first->prev->next = tt.first->prev->next ;
	    first->prev = tt.first->prev ;
	    tt.first->prev->next = first ;
	    tt.first->prev = hh ;
	  }
      }
    else
      first = tt.first ;
    tt.first = 0 ;
  }

        ///////////////////////////////////////////
        //
        // Dismembers the list, leaving the elements
        // pointed to by the list intact.
        //
  void clear () {
    dlink<K, T>* l = first ;
#ifdef DEBUG_LISTS
    printf (" clearing list: 0x%08x\n", (unsigned long)first) ;
#endif
    if (l == 0) return ;
    do {
      dlink<K, T>* ll = l ;
#ifdef DEBUG_LISTS
      printf ("   list-clear: [0x%08x]: (0x%08x, 0x%08x)\n", 
	      (unsigned long)ll, (unsigned long)ll->next, (unsigned long)ll->e) ;
#endif
      l = l->next ;
      delete ll ;
    } while (l != first) ;
    first = 0 ;
  }

  // zap() deletes both the list structure and the items pointed to
  // by the list.

        ///////////////////////////////////////////
        //
        // Deletes both the list structure and the objects
        // within the list.
        //
  void zap () {
    dlink<K, T>* l = first ;
#ifdef DEBUG_LISTS
    printf (" zapping list: 0x%08x\n", (unsigned long)first) ;
#endif
    if (l == 0) return ;
    do {
      dlink<K, T>* ll = l ;
      l = l->next ;
#ifdef DEBUG_LISTS
      printf ("   list-zap: [0x%08x]: (0x%08x, 0x%08x)\n", 
	      (unsigned long)ll, (unsigned long)ll->next, (unsigned long)ll->e) ;
#endif
      delete ll->e ;
      delete ll ;
    } while (l != first) ;
    first = 0 ;
  }

        ///////////////////////////////////////////
        //
        // A rather rudimentary print function for
        // displaying the contents of a list
        // on stdout.
        //
  void print ()
  {
    int done = 0 ;
    dlink<K, T>* l = first ;
    printf ("0x%08x -> ", (unsigned long)l ) ;
    while (l && !done)
    {
      printf ("(0x%08x, 0x%08x, 0x%08x) ", 
	      (unsigned long)l->e, (unsigned long)l->prev, 
	      (unsigned long)l->next) ;
      l = l->next ;
      done = (l == first) ;
    }
  }

        ///////////////////////////////////////////
        //
        // Returns a non-zero value if the list is
        // empty, otherwise returns 0.
        //
        // [out] int  list empty predicate
        //
  int empty () { return first == 0 ; }

        ///////////////////////////////////////////
        //
        // Constructor for an empty list.
        //
  dlist ()
  {
    first = 0 ;
    iterators = 0 ;
  }

        ///////////////////////////////////////////
        //
        // Constructor for a singleton list.
        //
        //  [in] T*  pointer to initial list item
        //
  dlist (T* a) 
  { 
    first = new dlink<K, T> (a, 0, 0) ;
    first->next = first->prev = first ;
    iterators = 0 ;
  }

        ///////////////////////////////////////////
        //
        // Destructor, but does not destroy list items.
        // However, it must clean up the iterator list.
        //
  ~dlist ()
  { 
    clear() ;
    while (iterators)
      {
	iterators->cs = 0 ;
	iterators->ce = 0 ;
	iterators->mark = 0 ;
	iterators = iterators->thread ;
      }
  }

  int notequal (int x, int y)       { return x != y ; }
  int notequal (char* x, char* y)   { return (y) ? strcmp (x, y) : 0 ; }
  int notequal (float x, float y)   { return x != y ; }
  int notequal (double x, double y) { return x != y ; }

protected:
  dlink<K, T>* first ;
  dlist_iterator<K, T> *iterators ;
};


/*---------------------------------------------------------------------
CLASS
     dlist_iterator

     Scans the elements of a dlist

OVERVIEW
     An iterator class for scanning through the elements of a dlist

----------------------------------------------------------------------*/

template <class K, class T> class dlist_iterator {

friend class dlist<K,T> ;

public:

        ///////////////////////////////////////////
        //
        // Constructor for empty list iterator.
        //
  dlist_iterator()
  { 
    cs = 0 ;
    ce = 0 ;
    mark = 0 ;
    thread = 0 ;
  }

        ///////////////////////////////////////////
        //
        // Constructor for initialised list iterator.
        // Use this to begin iterating over the list
        // given as parameter.
        //
        // The current position and the mark are both
        // set to the first item in the list.
        //
        //  [in] dlist&  a dlist to be iterated over
        //
  dlist_iterator(dlist<K, T>& s)
  { 
    cs = &s ;
    ce = cs->first ;
    mark = ce ;
    thread = 0 ;
    addToIteratorList (cs) ;
  }

        ///////////////////////////////////////////
        //
        // Constructor for initialised list iterator.
        // Use this to begin iterating over the list
        // given as parameter.
        //
        // The current position and the mark are both
        // set to the first item in the list.
        //
        //  [in] dlist*  an dlist that you want to iterate
        //
  dlist_iterator(dlist<K, T>* s)
  {
    cs = s ;
    ce = cs ? cs->first : 0 ;
    mark = ce ;
    addToIteratorList (cs) ;
  }
  
        ///////////////////////////////////////////
        //
        // Destructor for dlist_iterator must also
        // remove itself from the iterator list for
        // the list to which it is attached.
        //
  ~dlist_iterator()
  {
    if (cs)
      removeFromIteratorList (cs) ;
  }

        ///////////////////////////////////////////
        //
        // Re-assignment of an existing list iterator.
        //
        //  [in] dlist& a dlist to be iterated over
        //
  void operator= (dlist<K, T>& s)
  {
    if (cs)
      removeFromIteratorList (cs) ;
    cs = &s ;
    ce = cs->first ;
    mark = ce ;
    addToIteratorList (cs) ;
  }

        ///////////////////////////////////////////
        //
        // Re-assignment of an existing list iterator.
        //
        //  [in] dlist& a dlist to be iterated over
        //
  void operator= (dlist<K, T>* s)
  {
    if (cs)
      removeFromIteratorList (cs) ;
    if (s) {
      cs = s; ce = mark = cs->first ;
    }
    else {
      cs = 0; ce = 0 ;
    }
    addToIteratorList (cs) ;
  }

        ///////////////////////////////////////////
        //
        // The () operator is the principal method of
        // advancing the iterator through the list.
        // It returns the next item in the list and
        // moves the read pointer past that item.
        //
        //  [out] T* ptr to the next list item, or zero
        //           if end of list is reached.
  T* operator()()
  {
    if (!this) return 0 ;
    if (!this->cs) return 0 ;

    T* ret ;
    if (ce)
      {
	ret = ce->e ;
	ce = ce->next ;
	if (ce == cs->first)
	  ce = 0 ;
      }
    else
      ret = 0 ;

    return ret ;
  }

        ///////////////////////////////////////////
        //
        // Returns a pointer to the list item that is
        // next in sequence to be scanned by the ()
        // operator. This permits a one item lookahead
        // so that the next item can be either iterated
        // over or removed.
        //
  T* head ()
  {
    if (! this || ! cs) return 0 ;
    
    return ce ? ce->e : 0 ;
  }

        ///////////////////////////////////////////
        //
        // Removes the next unscanned item in the list.
        // This may cause the list to become empty,
        // and the () operator to subsequently return zero.
        // Calls to this function can be safely interleaved with
        // calls to the () operator and the head() function.
        //
  void remove ()
  {
    T* v = pull() ;
  }

        ///////////////////////////////////////////
        //
        // Removes and returns the next unscanned
        // item in the list, or zero if the iterator
        // is currently pointing at the end of the list
        // or if the list is empty.
        //
        // This may cause the list to become empty,
        // and the () operator to subsequently return zero.
        // Calls to this function can be safely interleaved with
        // calls to the () operator and the head() function.
        //
  T* pull ()
  {
    if (! this || ! cs || ! ce) return 0 ;

    dlink<K,T>* tmp = ce ;
    tmp->next->prev = tmp->prev ;
    tmp->prev->next = tmp->next ;

    if (tmp == cs->first)
      { // removing head of list
	if (tmp->next == tmp)
	  { // the list is now empty
	    cs->first = 0 ;
	    ce = mark = 0 ;
	  }
	else
	  ce = cs->first = cs->first->next ;
      }
    else
      { // Removing item from middle or end of list
	if (tmp == cs->first->prev)
	  ce = 0 ;         // removing last item, so stop scanning
	else
	  ce = tmp->next ; // in middle, so point at next unscanned
      }
    
    if (mark == tmp)
      mark = 0 ;
    tmp->next = tmp->prev = 0 ;
    T* v = tmp->e ;
    syncItemRemoval (tmp) ;
    delete tmp ;
    return v ;
  }


        ///////////////////////////////////////////
        //
        // Removes and returns the next item from
        // the "Mark" position, moving the mark
        // position on to the next item.
        // This function returns 0 if the mark is not
        // pointing at a valid list node, or if the mark
        // is pointing at the current iterator position.
        //
        // This may cause the mark to point at the end
        // of the list.
        //
  T* pull_mark ()
  {
    if (! this || ! cs || ! mark || mark == ce) return 0 ;

    dlink<K,T>* tmp = mark ;
    tmp->next->prev = tmp->prev ;
    tmp->prev->next = tmp->next ;

    if (tmp == cs->first)
      { // removing head of list
	if (tmp->next == tmp)
	  { // the list is now empty
	    cs->first = 0 ;
	    ce = mark = 0 ;
	  }
	else
	  mark = cs->first = cs->first->next ;
      }
    else
      { // Removing item from middle or end of list
	if (tmp == cs->first->prev)
	  mark = 0 ;         // removing last item, so stop scanning
	else
	  mark = tmp->next ; // in middle, so point at next unscanned
      }
    
    tmp->next = tmp->prev = 0 ;
    T* v = tmp->e ;
    syncItemRemoval (tmp) ;
    delete tmp ;
    return v ;
  }

        ///////////////////////////////////////////
        //
        // Inserts a new item into the list at a
        // position immediately preceeding the next
        // item to be scanned.
        //
        //  [in] T*  ptr to item to be added
        //
  void insert (T* a)
  {
    if (! this || ! cs) return ;

    if (ce)
      {
        dlink<K,T> *f = new dlink<K,T> (a, ce->prev, ce) ;
	ce->prev->next = f ;
	ce->prev = f ;
	if (ce == cs->first)
	  cs->first = f ;
      }
    else
      cs->append (a) ;
  }

        ///////////////////////////////////////////
        //
        // Inserts a new item into the list at a
        // position immediately preceeding the next
        // item to be scanned.
        //
        //  [in] T*  ptr to item to be added
        //
  void insert (T& a)
  {
    insert (&a) ;
  }

        ///////////////////////////////////////////
        //
        // Inserts a new item into the list at a
        // position immediately preceeding the Mark,
        // or at the end of the list if Mark is not
        // defined.
        //
        //  [in] T*  ptr to item to be added
        //
  void insert_mark (T* a)
  {
    if (! this || ! cs) return ;

    if (mark)
      {
        dlink<K,T> *f = new dlink<K,T> (a, mark->prev, mark) ;
	mark->prev->next = f ;
	mark->prev = f ;
	if (mark == cs->first)
	  cs->first = f ;
      }
    else
      cs->append (a) ;
  }

        ///////////////////////////////////////////
        //
        // Rewinds the list iterator to the start of
        // the list.
        //
  void rewind ()
  {
    ce = cs ? cs->first : 0 ;
  }

        ///////////////////////////////////////////
        //
        // Moves the current list iterator position
        // to the start of the list
        //
  void goto_start () { rewind () ; }

        ///////////////////////////////////////////
        //
        // Moves the current position of the list
        // iterator to the end of the list.
        //
  void goto_end () { ce = 0 ; }

        ///////////////////////////////////////////
        //
        // Moves the current position of the list
        // iterator to point at the given argument node.
        // If the argument is not present in the list
        // the current position is set to the end of the
        // list.
        //
  void goto_node (T* node)
  {
    dlink<K, T>* p = cs->first ;
    
    while (p && p->e != node)
      {
	p = p->next ;
	if (p == cs->first)
	  p = 0 ;
      }
    ce = p ;
  }

        ///////////////////////////////////////////
        //
        // Remembers the current position of this
        // list iterator in the "Mark" variable.
        //
  void set_mark () { mark = ce ; }


        ///////////////////////////////////////////
        //
        // Searches for the given node in the list
        // and sets the mark at that point if found.
        // Otherwise Mark is set to null.
        //
  void set_mark (T* node) 
  {
    dlink<K, T>* p = cs->first ;
    
    while (p && p->e != node)
      p = p->next ;
    
    mark = p ;
  }

        ///////////////////////////////////////////
        //
        // Positions the iterator at the position
        // last given to the "Mark" variable.
        //
  void goto_mark () { ce = mark ; }

        ///////////////////////////////////////////
        //
        // Returns the list item stored at the position
        // given by the "Mark" variable, or 0 if the
        // mark is not pointing at a valid list item, or
        // is effectively at the end of the list.
        //
  T* get_mark () { return mark ? mark->e : 0 ; }

        ///////////////////////////////////////////
        //
        // Exchanges the "Mark" variable with the
        // current iterator position.
        //
  void exchange_mark () { dlink<K, T>* tmp = mark ; mark = ce ; ce = tmp ; }

        ///////////////////////////////////////////
        //
        // Removes a section of this iterator's list
        // and return a pointer to that new list.
        // The section removed is from the "Mark"
        // position up to, but not including, the
        // current position or the end of the list
        // (whichever is reached first). Thus, if Mark
        // is closer to the end of the list than the
        // current position, the section removed will
        // Be delimited by (Mark, End_Of_List), otherwise
        // it will be delimited by (Mark, Current).
        // If Mark is pointing at the end of the list,
        // or is not pointing to a valid position
        // in the list, then no cut operation occurs, and
        // an empty list is returned.
        //
  dlist<K, T>* cut ()
  {
    dlist<K, T>* nl = new dlist<K, T> () ;
    T* item ;

    while (item = pull_mark() )
      nl->append (item) ;

    return nl ;
  }

        ///////////////////////////////////////////
        //
        // The items in the argument dlist are copied
        // into this iterator's list immediately prior
        // to the next item to be scanned (i.e. at the
        // current position).
        //
        // Ownership of the argument list is NOT assumed
        // by paste(), and responsibility for deleting
        // the argument list remains with the caller.
        //
  void paste (dlist<K, T>& nl)
  {
    dlist_iterator<K, T> next_item (nl) ;
    T* item ;

    while (item = next_item())
      insert (item) ;
  }

        ///////////////////////////////////////////
        //
        // The items in the argument dlist are copied
        // into this iterator's list immediately prior
        // to the Mark position.
        //
        // Ownership of the argument list is NOT assumed
        // by paste(), and responsibility for deleting
        // the argument list remains with the caller.
        //
  void paste_mark (dlist<K, T>& nl)
  {
    dlist_iterator<K, T> next_item (nl) ;
    T* item ;

    while (item = next_item())
      insert_mark (item) ;
  }

private:

  void printIteratorList ()
  {
    int count = 0 ;
    dlist_iterator<K,T> *di = cs ? cs->iterators : 0 ;
    printf ("dlist [0x%08x] ", cs) ;
    while (di && count < 10)
      {
	printf ("0x%08x ", di) ;
	++count ;
	di = di->thread ;
      }
    printf ("\n") ;
  }

  void addToIteratorList (dlist<K,T> *dl)
  {
    if (!dl)
      return ;

    // printf ("Adding iterator 0x%08x to \n", this) ;
    // printIteratorList () ;

    thread = dl->iterators ;
    dl->iterators = this ;

    // printf ("List is now: ") ;
    // printIteratorList () ;
  }

  void removeFromIteratorList (dlist<K,T> *dl)
  {
    if (!dl)
      return ;

    // printf ("Removing iterator 0x%08x from list \n", this) ;
    // printIteratorList () ;

    dlist_iterator<K,T> **di = &(dl->iterators) ;
    int done = 0 ;

    while (*di && !done)
      if (*di == this)
	{
	  *di = (*di)->thread ;
	  done = 1 ;
	}
      else
	di = &((*di)->thread) ;

    thread = 0 ;

    // printf ("List is now: ") ;
    // printIteratorList () ;
  }
  
  void syncItemRemoval (dlink<K,T> *v)
  {
    if (!cs) return ;
    dlist_iterator<K,T> *di = cs->iterators ;
    while (di)
      {
	if (di->ce == v)
	  di->ce = ce ;
	if (di->mark == v)
	  di->mark = mark ;
	di = di->thread ;
      }
  }

  dlink<K, T>* ce ;
  dlist<K, T>* cs ;
  dlink<K, T>* mark ;
  dlist_iterator<K, T>* thread ;
};

// This is a very simple stack template class

/*---------------------------------------------------------------------
CLASS
     stack

     Stack template class

OVERVIEW
     Provides a container for arbitary objects (of same type) with
     stack-oriented access protocols.
     Error detection and recovery conditions are minimal at present.

----------------------------------------------------------------------*/

template <class T> class stack : slist<int, T> {
public:
  stack () : slist<int, T> () { items = 0 ;}
  stack (stack<T> &s) : slist<int, T> (s) { items = length() ; }
  ~stack () {}
  void push (T& v) { prepend (&v) ; ++items ;}
  void push (T* v) { prepend (v) ; ++items ;}
  T* pop () { if (!items) underflow(); return get() ; }
  bool is_empty () { return (bool)(hd() == 0) ; }
  T* tos () { return hd() ; }
  
private:
  int items ;
  void underflow() { printf ("*** STACK UNDERFLOW ***\n") ; }
};

#endif // _lists_h_
