/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

// Header file for unparsing of parse trees back to source text.

#if ! defined _unparse_h_
#define _unparse_h_

#include "Basic.h"
#include "Strings.h"
#include "Lists.h"
#include "Defs.h"

#define LINE_BUFF_SIZE 13200
#define FORTRAN_LINE_LENGTH 72

     ////////////////////////
     //
     // An enumerated type to represent different language dialects
     //
typedef enum {
  ANSII_C,
  ANSII_Cpp,
  ANSII_F77, 
  CRAY_F77,
  CVX_F77, 
  CRAY_F90 
} dialect ;

     ////////////////////////
     //
     // An enumerated type to differentiate output formats
     //
typedef enum {
  PLAIN_TEXT,
  HTML_TEXT
} outputFormat ;

// Forward class declarations

class TypeStack ;
class SymbolStack ;
class SourceListing ;
class LineBuffer ;

// Non-member support functions

void CcUnparseStack (SourceListing&, TypeStack&, SymbolStack&) ;
void FortranUnparseExpression (LineBuffer&, PTR_EXPR) ;

// CLASS GROUP: classes to support unparsing of parse trees

/*---------------------------------------------------------------------
CLASS
     LineBuffer

     A buffer into which lines of unparsed text are deposited.

OVERVIEW
     A support class for unparsers. It provides a buffer into which
     text can be deposited during unparsing of parse tree structures.
     Normally a user would not need to use this class directly.

DESCRIPTION

     The LineBuffer class contains a character buffer into which
     characters can be placed during the composition of a single
     line of unparsed output. In addition to the text of the line
     there is a PTR_STMT pointer to the Statement which generated
     this line of text.
----------------------------------------------------------------------*/

class LineBuffer {
friend class SourceListing ;
  
  int isComment ;
  int isDirective ;
  int label ; 
  int tabs ;
  PTR_STMT stmt ;
  char buff [LINE_BUFF_SIZE] ;
  int pos ;

public:
  
     ///////////////////////
     //
     // Constructor.
     //
  LineBuffer () { pos = 0 ; }
  
     ///////////////////////
     //
     // When called, this function appends a null terminator to the
     // end of the buffer and returns a pointer to the start of text.
     //
  char* Text () { buff [pos++] = '\0' ; return buff ; }
  
     ///////////////////////
     //
     // Returns an integer equal to the number of characters in
     // the line buffer.
     //
  int Size ()   { return pos ; }

  
     ///////////////////////
     //
     // Provided the buffer size has not been exceeded, this operator
     // appends the given character to the line buffer. Otherwise
     // overflow is silently ignored.
     //
  void operator<< (char c) {
    if (pos < LINE_BUFF_SIZE)
      buff [pos++] = c ;
  }

  
     ///////////////////////
     //
     // This operator
     // appends the given string to the line buffer.
     // Non-printing characters are expanded using
     // standard C escape sequences.
     // Buffer overflow is not currently detected.
     //
  void operator<< (char *s) {
    for (char *p = s; p && *p; p++)
      switch (*p)
	{
	case '\0': *(this) << '\\' ; *(this) << '0' ; break ;
	case '\a': *(this) << '\\' ; *(this) << 'a' ; break ;
	case '\b': *(this) << '\\' ; *(this) << 'b' ; break ;
	case '\t': *(this) << '\\' ; *(this) << 't' ; break ;
	case '\n': *(this) << '\\' ; *(this) << 'n' ; break ;
	case '\v': *(this) << '\\' ; *(this) << 'v' ; break ;
	case '\f': *(this) << '\\' ; *(this) << 'f' ; break ;
	case '\r': *(this) << '\\' ; *(this) << 'r' ; break ;
	case '\\': *(this) << '\\' ; *(this) << '\\'; break ;
	default:
	  {
	    if (*p < 32)
	      {
		char sbuf [64] ;
		sprintf (sbuf, "%03o", *p) ;
		*(this) << '\\' ;
		*(this) << sbuf ;
	      }
	    else
	      *(this) << *p ;
	    break ;
	  }
	}
  }

     ///////////////////////
     //
     // This member function 
     // appends the given string to the line buffer.
     // Non-printing characters are not expanded.
     // Buffer overflow is not currently detected.
     //
  void accept_raw (char *s) {
    for (char *p = s; p && *p; p++)
      *(this) << *p ;
  }

  
     ///////////////////////
     //
     // Returns the previous character entered into the buffer, or
     // the null character if the buffer is empty.
     //
  char LastChar () { return (pos == 0) ? '\0' : buff [pos-1] ; }

  
     ///////////////////////
     //
     // Sets up the labeling, indentation and associated Statement
     // for a line of source text.
     //
     //  [in] int       label
     //  [in] int       number of tabs
     //  [in] int       non-zero if line is a comment line
     //  [in] int       non-zero if line is a directive
     //  [in] PTR_STMT  pointer to Statement at this line
     //
  void PrefixLine (int, int, int, int, PTR_STMT) ;
  

     ///////////////////////
     //
     // Sets up the indentation and associated Statement
     // for a line of C or C++ source text.
     //
     //  [in] int         number of tabs
     //  [in] Statement*  pointer to Statement at this line
     //  [in] CcLabel*    pointer to (optional) statement label
     //
  void CcStartLine (int, PTR_STMT, PTR_CC_LABEL=0) ;
  
     ///////////////////////
     //
     // Inserts the given number of tabs in the LineBuffer
     // at the current insertion point.
     //
     //  [in] int       the number of tabs to insert
     //
  void TabIndent (int) ;
};

/*---------------------------------------------------------------------
CLASS
     SourceLine

     Container class for a single line of unparsed source text.

OVERVIEW

     The SourceLine class is derived publicly from the string class,
     and in addition it contains a line number, and a pointer to the
     Statement from which the line of text is unparsed.

DESCRIPTION

     A SourceLine is normally created after a Statement has been
     unparsed into a LineBuffer object. Hence, the constructor for a
     SourceLine is passed a reference to a LineBuffer, from which
     it obtains its text.
----------------------------------------------------------------------*/

class SourceLine : public string {
public:

  // GROUP: public data members

  /////////////////////////////
  //
  // Contains the source line number.
  //
  int lineNumber ;

  /////////////////////////////
  //
  // Points to the Statement which originated this
  // unparsed source text.
  //
  PTR_STMT stmt ;

  /////////////////////////////
  //
  // Returns the line number of this SourceLine.
  //
  int index () { return lineNumber ; }

  /////////////////////////////
  //
  // Constructor.
  //
  // [in] LineBuffer&  reference to the source buffer
  // [in] PTR_STMT     pointer to Statement
  // [in] int          source line number
  //
  SourceLine (LineBuffer& lb, PTR_STMT ps, int l) : string (lb.Text())
  { 
    lineNumber = l ;
    stmt = ps ;
  }
};

/*---------------------------------------------------------------------
CLASS
     SourceListing

     A complete unparsed listing.

OVERVIEW

     A SourceListing object contains a list of SourceLine objects and
     thus represents an unparsed list of Statements.

DESCRIPTION

     The SourceListing class is derived publicly from an slist indexed
     by integer and yielding SourceLine objects, and also from the
     LineBuffer class. Each call to the Unparse function will create the
     contents of a SourceListing object from a given PTR_STMT pointer.
     Various other unparsing functions are also supported.

----------------------------------------------------------------------*/

class SourceListing : public slist<int, SourceLine>, public LineBuffer {

friend class SourceListingIterator ;

public:
  
  /////////////////////////////
  //
  // Returns the number of SourceLines of text in the listing
  //
  int sizeOfListing () ;
  
  /////////////////////////////
  //
  // Calls the appropriate unparser given a pointer to a Statement
  //
  //  [in] PTR_STMT  pointer to a Statement
  //  [in] dialect   enumeration of the dialect being unparsed
  //
  void unparse (PTR_STMT, dialect) ;
  
  /////////////////////////////
  //
  // Appends a discrete LineBuffer object to this SourceListing
  //
  //  [in] LineBuffer&  reference to the LineBuffer to be appended
  //
  void appendBuffer (LineBuffer&) ;
  
  /////////////////////////////
  //
  // Appends the current LineBuffer object to this SourceListing
  //
  void appendBuffer () ;
  
  /////////////////////////////
  //
  // Appends textual output to the current LineBuffer and returns
  // a reference to this SourceListing object so that the return
  // value of the expression involving a call to this operator can
  // be used as a the left-hand (object) operand in a subsequent
  // call to the << operator.
  // 
  //  [in] char*           string to be appended
  //  [out] SourceListing& this object
  //
  SourceListing& operator<< (char *s) {
    LineBuffer::operator<< (s) ;
    return *this ;
  }
  
  /////////////////////////////
  //
  // Appends textual output to the current LineBuffer and returns
  // a reference to this SourceListing object so that the return
  // value of the expression involving a call to this operator can
  // be used as a the left-hand (object) operand in a subsequent
  // call to the << operator.
  // 
  //  [in] char*           string to be appended
  //  [out] SourceListing& this object
  //
  SourceListing& operator<< (char c) {
    LineBuffer::operator<< (c) ;
    return *this ;
  }
  

  /////////////////////////////
  //
  // Constructor. Sets default source language to F77
  //
  SourceListing () : slist<int, SourceLine> () { lang = F77 ; ofmt = PLAIN_TEXT ; }
  
  /////////////////////////////
  //
  // Constructor, for a given source language.
  //
  SourceListing (language l, outputFormat o) : slist<int, SourceLine> ()
    {
      lang = l ;
      ofmt = o ;
    }
  
  /////////////////////////////
  //
  // Destructor.
  //
  ~SourceListing () ;

private:
  language lang ;
  outputFormat ofmt ;
} ;

/*---------------------------------------------------------------------
CLASS
     SourceListingIterator

     For iterating through a SourceListing

OVERVIEW
     Create one of these objects if you want to iterate through a
     SourceListing. Please consult the slist_iterator template for
     details on how to use this class.
----------------------------------------------------------------------*/

class SourceListingIterator : public slist_iterator<int, SourceLine> {

friend class SourceListing ;

public:

  ////////////////////////////////////////
  //
  // Constructor.
  //
  //  [in] SourceListing&  reference to the SourceListing to be iterated through.
  //
  SourceListingIterator (SourceListing& sl) : slist_iterator<int, SourceLine> (sl) { }
};

#endif // _unparse_h_
