//
// Class CSIParse for simple instruction parsing
//
// Copyright (C) 2000-2004 by Grigori G.Fursin
//
// http://homepages.inf.ed.ac.uk/gfursin
//
// ICSA, School of Informatics,
// University of Edinburgh, UK
//

//Import libraries...
import java.io.*;

public class CSIParse
{
  //Variables
  private int XNumOptMax;
  private int XIgnoreCase;
  
  private String[] Instr;
  private int[]    HasParam;
  private char[] Sep;
  private String[] Param;
  private String[] DefParam;

  private int CurParam;

  private String inst;
  private String prm;
    
  //****************************************************************************
  //"Initialization"
  //
  //Input:  NumOptMax   = Max Number of Options
  //        IgnoreCase  = 1 - Ignore case, 0 - Don't ignore case
  //
  //Return: 0  - Ok
  //        >0 - Error
  //

  public int Init(int NumOptMax, int IgnoreCase)
  {
    int x;

    XNumOptMax=NumOptMax;
    XIgnoreCase=IgnoreCase;

    if (NumOptMax<1) return 1;

    if ((IgnoreCase!=0) && (IgnoreCase!=1)) return 1;

    Instr=new String[XNumOptMax];
    HasParam=new int[XNumOptMax];
    Sep=new char[XNumOptMax]; //separator if parameter exist
    Param=new String[XNumOptMax];

    for (x=0; x<XNumOptMax; x++) 
    {
      Instr[x]="";
      HasParam[x]=0;
      Sep[x]=0;
      Param[x]="";
    }

    CurParam=0;

    //temporal
    inst="";
    prm="";

    return 0;
  }

  //****************************************************************************
  //"Add Instruction"
  //
  //Input:  Inst    = Instruction
  //        HasPrm  = Has Parameter?
  //        Separ   = Separator for parameter
  //        Default = Default parameter
  //
  //Return: 0  - Ok
  //        >0 - Error
  //
  public int Add(String Inst, int HasPrm, char Separ, String Default)
  {
    if (CurParam>XNumOptMax) return 1;

    Instr[CurParam]=Inst;
    HasParam[CurParam]=HasPrm;
    Sep[CurParam]=Separ;
    Param[CurParam]=Default;

    if (XIgnoreCase==1)
    {
      Instr[CurParam]=Instr[CurParam].toLowerCase();
    }

    CurParam++;

    return 0;
  }

  //****************************************************************************
  //"Parse Command Line"
  //
  //Return: <0  - Error
  //        >=0 - Identificator 
  //
  public int Parse(String str)
  {
    int nPos=0;
    int nLen=str.length();

    if (nLen==0) return -1; //Line is empty

    //ignoring first spaces
    while (str.charAt(nPos)==' ')
    {
      nPos++;
      if (nPos>=nLen) return -1; //Line is empty
    }

    //getting instruction until end/separator/' '
    inst="";

    int found1=0;
    while (found1==0)
    {
      if (!((str.charAt(nPos)>='a') && (str.charAt(nPos)<='z')) && 
          !((str.charAt(nPos)>='A') && (str.charAt(nPos)<='Z')) && 
          !((str.charAt(nPos)>='0') && (str.charAt(nPos)<='9')) && 
          (str.charAt(nPos)!='_')) 
      {
        found1=1;
      }
      else
      {
        inst+=str.charAt(nPos);

        nPos++;

        if (nPos>=nLen) found1=1;
      }
    }

    if (XIgnoreCase==1) inst=inst.toLowerCase();

    //Search for instruction
    found1=0;
    int i;
    for (i=0; i<CurParam; i++)
    {  
      if (Instr[i].equals(inst))
      {
        found1=1;
        break;
      }
    }

    if (found1==0) return -2; // Unknown instruction

    if (nPos>=nLen)
    {
      if (HasParam[i]==0) return i; //Return instruction
      return -3; //Instruction needs parameter
    }

    //ignoring spaces
    int foundspaces=0;
    while (str.charAt(nPos)==' ')
    {
      foundspaces=1;
      nPos++;
      if (nPos>=nLen) 
      {
        if (HasParam[i]==0) return i; //Return instruction
        return -3; //Instruction needs parameter
      }
    }

    if (HasParam[i]==0) return -4; //Instruction doesn't need parameters

    if (Sep[i]==' ')
    {
      if (foundspaces==0) return -5; //Wrong separator
    }
    else
    {
      if (Sep[i]!=str.charAt(nPos)) return -5; //Wrong separator
      nPos++;
    }

    if (nPos>=nLen) return -3; //Instruction needs parameter

    //ignoring spaces
    while (str.charAt(nPos)==' ')
    {
      nPos++;
      if (nPos>=nLen) return -3; //Instruction needs parameter
    }

    //getting parameter until end/' '
    prm="";

    while (nPos<nLen)
    {
      prm+=str.charAt(nPos);

      nPos++;
    }
  
    Param[i]=prm.trim();

    return i; //Return instruction
  }

  //****************************************************************************
  //"Get String Parameter"
  //
  public String GetParamStr(int NumParam)
  {   
    if ((NumParam<0) || (NumParam>=CurParam)) return null;

    return Param[NumParam];
  }

  //****************************************************************************
  //"Get LowerCase String Parameter"
  //
  public String GetParamStrLwr(int NumParam)
  {   
    if ((NumParam<0) || (NumParam>=CurParam) || (HasParam[NumParam]==0)) return null;

    return Param[NumParam].toLowerCase();
  }

  //****************************************************************************
  //"Get Long Parameter"
  //
  public long GetParamLong(int NumParam)
  {
    if ((NumParam<0) || (NumParam>=CurParam) || (HasParam[NumParam]==0)) return 0;
 
    long ln=0;
    try
    {
      ln=Long.valueOf(Param[NumParam]).longValue();
    }
    catch (NumberFormatException nfe1)
    {}

    return ln;
  }

  //****************************************************************************
  //"Get Int Parameter"
  //
  public int GetParamInt(int NumParam)
  {
    if ((NumParam<0) || (NumParam>=CurParam) || (HasParam[NumParam]==0)) return 0;
  
    int in=0;
    try
    {
      in=Integer.valueOf(Param[NumParam]).intValue();
    }
    catch (NumberFormatException nfe1)
    {}
    
    return in;
  }

}
