/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

#if ! defined _scope_h_
#define _scope_h_

#include "Defs.h"
#include "Symbols.h"

// #define TRACE_SYMTAB	1

#define MAX_SCOPES	256
#define CASE_SENS	0
#define NOT_CASE_SENS	1

/*--------------------------------------------------------------

CLASS
    ScopeStack

    Implements a stack of symbol scopes during parsing.

DESCRIPTION

 This class manages scoping levels during parsing through a
 stack abstraction. At any given point during compilation the
 parser will be at a given static nesting level. The path from
 that level to the outermost level can be represented by a 
 stack of scope levels. The parser simply arranges to push and
 pop scope levels at appropriate times, according to the
 syntactic entities recognised.

 Note: this class is really only useful during parsing of source
 code, and has no useful function once a parse tree has been
 successfully built.

----------------------------------------------------------------*/

class ScopeStack {

private:

  PTR_SCOPE *scope_stack ;
  int top_scope_level ;
  int top_scope_start ;
  int old_level ;
  PTR_STMT current_procedure ;

public:

    ////////////////////////////////////
    //
    // Constructor.
    //
  ScopeStack (int depth = MAX_SCOPES) {
    int i ;
    scope_stack = new PTR_SCOPE [depth] ;
    for (i = 0; i < depth; i++)
      scope_stack [i] = 0 ;
    top_scope_level = top_scope_start = 0 ;
    current_procedure = 0 ;
  }

    ////////////////////////////////////
    //
    // Destructor.
    //
  ~ScopeStack () { 
    delete scope_stack ;
  }

    ////////////////////////////////////
    //
    // [out] ScopeNode* returns current scope
    //
  PTR_SCOPE cur_scope ()          { return scope_stack[top_scope_level] ; }

    ////////////////////////////////////
    //
    // [out] ScopeNode* returns the global scope
    //
  PTR_SCOPE global_scope ()       { return scope_stack[1] ; }

    ////////////////////////////////////
    //
    // Pushes a new scope for storing function arguments prior
    // to reducing the function declaration. These are later
    // copied into the scope of the function using the
    // constructor for a ScopeNode.
    //
  void push_arg_scope () ;

    ////////////////////////////////////
    //
    // Pops a scope from the ScopeStack.
    //
  void pop_normal_scope () ;

    ////////////////////////////////////
    //
    // [out] Statement* returns a pointer to the current procedure.
    //
    //
  PTR_STMT cur_proc ()            { return current_procedure ; }

    ////////////////////////////////////
    //
    // Sets the current procedure. This is normally only used
    // during parsing.
    //
    // [in] Statement* the current procedure
    //
  void set_current_proc (PTR_STMT p) { current_procedure = p ; }

    ////////////////////////////////////
    //
    // Converts the given Symbol to a global entity, effectively
    // moving it from the current level to the top level.
    //
    // [in] Symbol* pointer to the symbol to be moved
    //
  void convert_to_global (PTR_SYMB) ;

    ////////////////////////////////////
    //
    // Returns the parent scope of a numbered scope level.
    // The global scope is level number 1.
    //
    // [in]  int        scope level to find parent of.
    // [out] ScopeNode* pointer to parent scope
    //
  PTR_SCOPE parent_scope (int) ;

    ////////////////////////////////////
    //
    // Pushes the scope level and creates a new, empty
    // scope.
    //
    // [in] Statement* the defining statement for the new scope
    // [in] int        
    //
  void push_scope_level (PTR_STMT, int) ;

    ////////////////////////////////////
    //
    // Pops a scope level for a given statement
    //
  void pop_scope_level (PTR_STMT) ;

    ////////////////////////////////////
    //
    // Pushes a given scope and makes it the
    // current scope. This is primarily intended
    // for use with C++ classes and namespaces,
    // which have their own scope.
    //
    // [in] ScopeNode* an existing ScopeNode
    //
  void push_scope (PTR_SCOPE) ;

    ////////////////////////////////////
    //
    // Pops the current scope from the ScopeStack,
    // and returns to the outer (or previous) scope
    //
  PTR_SCOPE pop_scope () ;

    ////////////////////////////////////
    //
    // Returns the integer assigned to the current
    // scope level
    //
  int cur_scope_level () ;

    ////////////////////////////////////
    //
    // Returns the ScopeNode given by the integer
    // nest level given as a parameter.
    //
  PTR_SCOPE scope_at_level (int) ;

    ////////////////////////////////////
    //
  PTR_STMT cur_parent () ;
};

#define DEFAULT_SCOPE_SIZE 57

#define LOOK_LOCALLY	 0
#define LOOK_GLOBALLY    1
#define SEARCH_ONLY      0
#define SEARCH_OR_CREATE 1
#define NO_MATCH	 0


/*--------------------------------------------------------------

CLASS
    ScopeNode

    Container for symbols within a local scope

DESCRIPTION

 Each symbol scope in the source code has an associated ScopeNode
 object. This is attached to the Statement or Symbol node which defines
 the bounds of that scope, and is accessible via the scoping() function
 provided by the Statement class. Block or function scopes are 
 attached to their respective Statement nodes. Class scopes
 are attached to the defining class symbol - which is itself 
 contained in some scope. Namespace scopes are very similar to
 class scopes, but are attached to the namespace symbol which
 defines them. The unamed scope is an alias for the global file
 scope, which also has a symbol.

 A scope is actually a local Symbol table with some associated
 information. This includes pointers to the Statement which
 defines the scope, a pointer to the enclosing scope (i.e.
 the parent scope), and a printable name for the scope.
 
 As an individual Symbol table has a dynamically variable size
 (it is implemented ultimately as a Map object), there is no
 space inefficiency problem caused by having large numbers of
 scopes attached to a parse tree.

----------------------------------------------------------------*/

class ScopeNode {

friend class ScopeStack ;
friend class ScopeIterator ;

public:

    ////////////////////////////////////
    //
    // Constructor for function body scopes which inherits arguments
    // as local names.
    //
    // [in] ScopeNode& function argument Symbols to preload
    // [in] Statement* the defining Statement
    // [in] ScopeNode* the enclosing scope
    // [in] char*      the scope name
    //
  ScopeNode (ScopeNode& preloads, PTR_STMT h, PTR_SCOPE e = 0, const char *n = 0) 
  : name (n), declared (preloads.declared) { 
    header = h ; 
    enclosing = e ; 
    scope_symb = 0 ;
  }
  
    ////////////////////////////////////
    //
    // Constructor for block scopes
    //
    // [in] Statement* the defining Statement
    // [in] ScopeNode* the enclosing scope
    // [in] char*      the scope name
    //
  ScopeNode (PTR_STMT h, PTR_SCOPE e = 0, const char *n = 0) 
  : name (n) { 
    header = h ; 
    enclosing = e ;
    scope_symb = 0 ;
  }
  
    ////////////////////////////////////
    //
    // Constructor for class or namespace scopes. Note the
    // first argument must be a valid Symbol pointer.
    //
    // [in] Symbol*    non-null class or namespace symbol for this scope
    // [in] ScopeNode* pointer to the enclosing scope
    // [in] char*      the scope name
    //
  ScopeNode (PTR_SYMB s, PTR_SCOPE e = 0) 
  : name (s->ident) { 
    header = 0 ; 
    enclosing = e ;
    scope_symb = s ;
  }
  
    ////////////////////////////////////
    //
    // Destructor.
    //
  ~ScopeNode () { }

    ////////////////////////////////////
    //
    // Deletes all Symbols within this scope.
    //
  void zap () { declared.DeleteEntries () ; }

    ////////////////////////////////////
    //
    // [out] Statement* pointer to defining Statement
    //
  PTR_STMT head () { return header ; }

    ////////////////////////////////////
    //
    // In Fortran variables can be implicitly typed.
    // This function defines the rules for infering
    // type information from the initial character
    // of the symbol.
    //
    // [in] Datatype* [26]  array of pointers to types
    //
  void set_implicit (PTR_TYPE itype[26]) {
    declared.set_implicit (itype) ;
  }

    ////////////////////////////////////
    //
    // [out] ScopeNode* returns a pointer to the enclosing scope
    //
  PTR_SCOPE enclosing_scope () { return enclosing ; }

    ////////////////////////////////////
    //
    // [in] Symbol* symbol to be removed from this scope
    //
  void remove_symbol (PTR_SYMB s) {
    PTR_HASH h = declared [s->ident()] ;
    if (h)
    {
      declared.remove_variant (s->ident(), NO_MATCH) ;
      delete h ;
    }
  }

  // GROUP: symbol lookup and install functions

    ////////////////////////////////////
    //
    // The fundamental symbol table search routine.
    //
    // Search for a character string in the hash table
    // associated with this scope, and return a pointer
    // to the hash entry.
    //
    // Search my be local or global, may insist on a match
    // with a symbol variant, and may or may not create a
    // new table entry (but NOT a Symbol) if the given string is not in scope.
    //
    // [in]  char*       the characters of the symbol to find
    // [in]  int         LOOK_GLOBALLY or LOOK_LOCALLY
    // [in]  int         MATCH or NO_MATCH
    // [in]  int         SEARCH_ONLY or SEARCH_OR_CREATE
    // [in]  int         CASE_SENS or NOT_CASE_SENS
    // [out] hash_entry* pointer to symbol hash entry
    //
  PTR_HASH find (const char *, int, int, int, int=CASE_SENS) ;

    ////////////////////////////////////
    //
    // Looks for a symbol in the symbol table. If the
    // second argument is SOFT then the search is global,
    // otherwise it is purely local to this scope.
    // If no matching symbol is found, a new symbol is created.
    //
    // [in]  char*       the symbol characters
    // [in]  int         SOFT or HARD
    // [out] hash_entry* pointer to symbol hash entry
    //
  PTR_HASH look_up_ftn (const char *str, int decl_type) {
    if (decl_type == SOFT) 
      return find (str, LOOK_GLOBALLY, NO_MATCH, 
		   SEARCH_OR_CREATE, NOT_CASE_SENS) ;
    else
      return find (str, LOOK_LOCALLY, NO_MATCH, 
		   SEARCH_OR_CREATE, NOT_CASE_SENS) ;
  }

    ////////////////////////////////////
    //
    // Searches globally for a symbol and creates one if
    // no symbol with the same character string is found.
    //
    // [in]  char*       the symbol characters
    // [out] hash_entry* pointer to symbol hash entry
    //
  PTR_HASH look_up_ftn_sym (const char *str) { 
    return find (str, LOOK_GLOBALLY, NO_MATCH, SEARCH_OR_CREATE,
		 NOT_CASE_SENS) ; }

    ////////////////////////////////////
    //
    // Searches globally for a symbol but does not create
    // a new symbol if no matching symbol is found.
    //
    // [in]  char*       the symbol characters
    // [out] hash_entry* pointer to symbol hash entry
    //
  PTR_HASH just_look_up_ftn_sym (const char *str) {
    return find (str, LOOK_GLOBALLY, NO_MATCH, SEARCH_ONLY,
		 NOT_CASE_SENS) ; }

    ////////////////////////////////////
    //
    // Searches locally for a symbol but does not create
    // a new symbol if no matching symbol is found.
    //
    // [in]  char*       the symbol characters
    // [out] hash_entry* pointer to symbol hash entry
    //
  PTR_HASH just_look_up_ftn_sym_in_scope (const char *str) {
    return find (str, LOOK_LOCALLY, NO_MATCH, SEARCH_ONLY,
		 NOT_CASE_SENS) ; }

    ////////////////////////////////////
    //
    // Searches for an operator symbol.
    //
    // [in] int          integer id for the operator
    // [out] hash_entry* pointer to operator symbol hash entry
    //
  PTR_HASH look_up_op (int op) ;

    ////////////////////////////////////
    //
    // Installs a hash_entry object in this scope and sets the
    // scope pointer of the hash_entry object to point at this scope.
    //
    // [in] hash_entry*  pointer to hash_entry object
    //
  void install_entry (PTR_HASH h) { 
    declared += h ;
    h->set_attr (this) ;
  }

    ////////////////////////////////////
    //
  void dump (int) ;

    ////////////////////////////////////
    //
    // Places a textual description in the form A::B::C
    // for this scope, in the given buffer.
    //
    // [in] char** address of char buffer for output
    // [in] char*  lower limit of output buffer
    //
  void sprint (char**, char*) ;

    ////////////////////////////////////
    //
    // Returns a string describing the category of this scope
    // which will be either "class", "namespace", "function",
    // "function prototype", "local".
    //
    // [out] const char*  description of scope
    //
  const char* category () ;

    ////////////////////////////////////
    //
    // Returns the name of the scope
    //
    // [out] string&  the scope name
    //
  string& scope_name () { return name ; }

    ////////////////////////////////////
    //
    // Returns the name of the scope
    //
    // [out] string&  the scope name
    //
  PTR_SYMB scope_sym () { return scope_symb ; }

    ////////////////////////////////////
    //
    // Returns the number of symbols in the scope
    //
    // [out] int  the count of symbols
    //
  int num_symbols () { return declared.contents() ; }

    ////////////////////////////////////
    //
  SymbolTable& symbols() { return declared ; }

private:
  string      name ;       // each scope has a printable name
  PTR_STMT    header ;     // ptr to head-of-scope statement
  PTR_SYMB    scope_symb;  // ptr to scope name symbol (class or namespace)
  SymbolTable declared ;   // declared symbols at this level
  PTR_SCOPE   enclosing ;  // ptr to enclosing scope (or null)
};


class ScopeIterator : public SymbolTable_Iterator {
public:
  ScopeIterator (ScopeNode& s) : SymbolTable_Iterator (s.declared) {}
};

/*--------------------------------------------------------------

CLASS
    Scope_Kind

    Used to tag a ScopeRec object with the kind of scope that
    it defines.

DESCRIPTION

    In C++ there are four kinds of scope. These are enumerated
    by the scope_tag enum type, which is then enclosed in the
    Scope_Kind class to provide a suitable object wrapper and
    appropriate member functions on scope kinds.

---------------------------------------------------------------*/

class Scope_Kind {
public:
  scope_tag kind ;

  Scope_Kind () { kind = LOCAL_SCOPE ; }
  Scope_Kind (scope_tag k) { kind = k ; }
  Scope_Kind (Scope_Kind &k) { kind = k.kind ; }

  char *kname () { return kstrings [kind] ; }

private:
  static char* kstrings[] ;
};

#endif // _scope_h_
